// Written in the D Programming language.
/*
 * Sanspam: Mailbox utility to delete/bounce spam on server interactively.
 * Copyright (C) 2018  Dennis Katsonis dennisk@netspace.net.au
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

import std.typecons;
import std.conv;
import std.algorithm;
import std.string;
import std.regex;
import std.base64;
import std.encoding;
import std.array;

enum space = hexString!"20";

enum encodingLabelUTF8 = "utf-8";
enum encodingLabelISO_8859_1 = "iso-8859-1";
enum encodingLabelISO_8859_2 = "iso-8859-2";

alias textEncodingType = Tuple!(charsetType, "charset", encodingType, "encoding");

enum charsetType {
	CLEAR_TEXT,
	UTF8,
	ISO8859_1,
	ISO8859_2,
	UNKNOWN
}

enum encodingType {
	ASCII,
	BASE64
} 

textEncodingType getTextEncodingType(in string charset, in string encoding) @safe pure
{
	/* Function to determine the encoding type from supplied text.
	   This return the text Encoding Type, which is a tuple containing
	   both the character set used, and whether ASCII or BASE64
	   encoded. */
	textEncodingType te;

	if (charset.toLower == encodingLabelUTF8)
		te.charset = charsetType.UTF8;
	else if (charset.toLower == encodingLabelISO_8859_1)
		te.charset = charsetType.ISO8859_1;
	else if (charset.toLower == encodingLabelISO_8859_2)
		te.charset = charsetType.ISO8859_2;
	else
		te.charset = charsetType.UNKNOWN;

	if (encoding.toUpper == "Q")
		te.encoding = encodingType.ASCII;
	else if (encoding.toUpper == "B")
		te.encoding = encodingType.BASE64;
	else
		/* ASCII if we don't know */
		te.encoding = encodingType.ASCII;

	return te;
}


T hextoChar(T)(string input) @safe pure
	 if (is(T == char) || is(T == Latin1Char) || is(T == Latin2Char))
		 in {
			 assert(input.length == 2);
			 assert(input[0] >= '0');
			 assert(input[0] <= 'F');
			 assert(input[1] >= '0');
			 assert(input[1] <= 'F');
			 }
do {
	return cast(T)(parse!int(input,16));
 }

string base64Encode(string text) pure @trusted
{
	string output;
	return Base64.encode(cast(ubyte[])text);
}

string base64Decode(string text) pure
{
	string output;
	ubyte[] array = cast(ubyte[])text;
	ubyte[] decoded;
	try
		{
			decoded = Base64.decode(array);
		} catch (Base64Exception e) {
		return "";
	}
	foreach(x; decoded) {
		// Rebuild text as decoded characters;
		output~=x;
	}
	return output;
}

string decodeText(string text, textEncodingType te) 
{
	auto output = appender!string;
	int index;

	assert(index <= text.length);
	text = text[index..$];
	if (te.encoding == encodingType.BASE64) {
		try {
			text = text.base64Decode;
		} catch (Base64Exception e) {
			// Error decoding, return original string.
			return text;
		}
	}
	
	if (te.charset == charsetType.UTF8) {
		output.put(decodeUTF8!string(text));
	} else if (te.charset == charsetType.ISO8859_1) {
		output.put(decodeUTF8!Latin1String(text));
	} else if (te.charset == charsetType.ISO8859_2) {
		static if (is(Latin2String)) {
			// Some earlier compilers may not support Latin2String.
			// Only decode to Latin2String if possible, othewise default to UTF-8.
			output.put(decodeUTF8!Latin2String(text));
		} else {
			output.put(decodeUTF8!string(text));
		}
	} else if (te.charset == charsetType.UNKNOWN) {
		output.put(decodeUTF8!string(text)); // If not known, use UTF-8 and hope for the best.
	}
	
	return output[];
}


string decodeUTF8(T)(string text)
	 if (is(T == string) || is(T == Latin1String) || is(T == Latin2String)) {
		 Appender!T decodedText = appender!T;
		 string output;
		 int index;
		 while(index < text.length) {
			 if (text[index] == '=') {
				 if ((index + 3) > text.length) {
					 static if(is(T == string)) {
						 decodedText.put(text[index].to!string);
					 }
					 static if(is(T == Latin1String)) {
						 decodedText.put(text[index].to!Latin1Char);
					 }
					 static if(is(T == Latin2String)) {
						 decodedText.put(text[index].to!Latin2Char);
					 }
					 break;
				 }
				 static if(is(T == string)) {
					 string hchars = text[index+1..index+3];
					 decodedText.put(hextoChar!char(hchars));
				 }
				 static if (is(T == Latin1String)) {
					 string hchars = text[index+1..index+3].to!string;
					 decodedText.put(hextoChar!Latin1Char(hchars));
				 }
				 /* Some earlier compilers may not support
					Latin2String.  Only decode to Latin2String if
					possible, othewise default to UTF-8. */
				 
				 static if (is(Latin2String)) {
					 static if (is(T == Latin2String)) {
						 string hchars = text[index+1..index+3].to!string;
						 decodedText.put(hextoChar!Latin2Char(hchars));
					 }
					 
				 } else {
				 }
				 index+=3;
				 
			 } else if (text[index] == '_') {
				 static if(is(T == string)) {
					 decodedText.put(space);
				 }
				 static if(is(T == Latin1String)) {
					 T converted;
					 transcode(space.to!T, converted);
					 decodedText.put(converted);
				 }
				 static if (is(Latin2String)) {
					 static if(is(T == Latin2String)) {
						 T converted;
						 transcode(space.to!T, converted);
						 decodedText.put(converted);
					 }
				 }
				 index++;
				 
			 } else {
				 static if(is(T == string)) {
					 decodedText.put(text[index++]);
				 }
				 static if(is(T == Latin1String)) {
					 T converted;
					 try {
						 transcode(text[index++].to!string, converted);
					 } catch (EncodingException e) {
						 decodedText.put(text[index++].to!Latin1Char);
					 }
					 decodedText.put(converted);
				 }
				 
				 static if (is(Latin2String)) {
					 static if(is(T == Latin2String)) {
						 T converted;
						 try {
							 transcode(text[index++].to!string, converted);
						 } catch (EncodingException e) {
							 decodedText.put(text[index++].to!Latin2Char);
						 }
						 decodedText.put(converted);
					 }
				 }
			 }
		 } // end while
		 static if(is(T == string)) {
			 return decodedText[];
		 }
		 static if(is(T == Latin1String)) {
			 transcode(decodedText[], output);
			 return output;
		 }
		 static if (is(Latin2String)) {
			 static if(is(T == Latin2String)) {
				 transcode(decodedText[], output);
				 return output;
			 }
		 }
	 }

string convertText(string input)
{
	textEncodingType te;

	auto text = appender!string;
	string output;
	auto result = matchAll(input, regex(r"=\?.+?\?=($|\s)"));

	/* If no match, just return the input and do no conversion. */
	if (result.empty) {
		return input;
	}
	output~=result.pre; /* If there is anything before the first "=?", add this.
						  Sometimes subject lines contain unencoded information mixed
						  with encoded */

	foreach(item; result)
		{
			auto x = matchFirst(item.front, regex(r"=\?(?P<charset>.+?)\?(?P<encoding>.+?)\?(?P<subject>.+)\?="));
			/* Extract the character set and encoding type, as well as
			   the remaining text.  These will be positions 1, 2 and 3
			   in the regex match. */
			if(x.empty) {
				te.encoding = encodingType.ASCII;
				te.charset = charsetType.CLEAR_TEXT;
				text.put(item.front);
			} else {
				string converted;
				te = getTextEncodingType(x["charset"], x["encoding"]);
				
				if (te.encoding == encodingType.BASE64) {
					converted = x["subject"].base64Decode;
					if (converted == "") {
						te.charset = charsetType.CLEAR_TEXT;
						text.put(x["subject"]);
					} else {
						text.put(converted);
					}
				} else {
					text.put(x["subject"]);  // Append subject text
				}
			}
		} // end of foreach
	/* Now decode based on the character set that was determined earlier. */
	if ((te.charset == charsetType.CLEAR_TEXT) ||
		(te.charset == charsetType.UNKNOWN)) {
		/* If clear text or unknown, do nothing */
		output~=text[];
	} else if (te.charset == charsetType.UTF8) {
		output ~= decodeUTF8!string(text[]);
	} else if (te.charset == charsetType.ISO8859_1) {
		output ~= decodeUTF8!Latin1String(text[]);
	} else if (te.charset == charsetType.ISO8859_2) {
		static if (is(Latin2String)) {
			// Some earlier compilers may not support Latin2String.
			// Only decode to Latin2String if possible, othewise default to UTF-8.
			output ~= decodeUTF8!Latin2String(text[]);
		} else {
			output ~= decodeUTF8!string(text[]);
		}
	}
	return output;
}

unittest
{
	assert(hextoChar!char("E2") == 226);
	assert(hextoChar!char("10") == 16);
	assert(hextoChar!Latin1Char("10") == 16);
	static if (is(Latin2String)) {
		assert(hextoChar!Latin2Char("10") == 16);
	}
	assert(hextoChar!char("FF") == 255);
	assert(hextoChar!char("00") == 0);
	static if (is(Latin2String)) {
		assert(hextoChar!Latin2Char("00") == 0);
	}
	assert(hextoChar!Latin1Char("00") == 0);
	assert(getTextEncodingType("utf-8","Q").charset == charsetType.UTF8);
	assert(getTextEncodingType("utf-8","Q").encoding == encodingType.ASCII);
	assert(convertText("Plain text") == "Plain text");
	assert(convertText("=?Plain text?=") == "=?Plain text?=");
	assert(convertText("=?Plain? text?=") == "=?Plain? text?=");
	assert(convertText("Plain=? text?=") == "Plain=? text?=");
	assert(getTextEncodingType("uTF-8","B").charset == charsetType.UTF8);
	assert(getTextEncodingType("uTF-8","B").encoding == encodingType.BASE64);
	assert(convertText("=?utf-8?Q?hello?=") == "hello");
	assert(convertText("Subject: =?ISO-8859-1?B?SWYgeW91IGNhbiByZWFkIHRoaXMgeW8=?=") == "Subject: If you can read this yo");
	assert(convertText("Subject: =?utf-8?B?F}FFFAIL=?=") == "Subject: F}FFFAIL=");
	assert(convertText("Subject: =?ISO-8859-1?B?F}FFFAIL=?=") == "Subject: F}FFFAIL=");
	assert(convertText("=?UTF-8?B?QUJDIFN0b3JlOiBOZXcgT3JkZXI=?=") == "ABC Store: New Order");
	assert(convertText("=?UtF-8?Q?=F0=9F=98=A1British_politicians_=27cover?= =?utf-8?Q?ed_up_child_abuse_for_decades=27?=") == "😡British politicians 'covered up child abuse for decades'");
	assert(convertText("=?UtF-8?Q?=F0=9F=98=A1British_politicians_=27cover?= =?utf-8?Q?ed_up_child_abuse_for_decades=?=") == "😡British politicians 'covered up child abuse for decades=");
}

